private import common;
private import scene_select;

private import std.thread;
private import std.conv;
private import std.string;

class ReplayEntry {
    int gameScore;
    int turnScore;
    int clearScore;
    int leftScore;
    int nomissScore;
    int allScore;
    long utcTime;
    int patterns;

    int endTurn;
    int[] scoreStat;
    enum ScoreStatEnc {
        CSV, DIV_PRINTABLE95
    }
    ScoreStatEnc scoreStatEnc;

    char[] mode;
    long seed;
    char[] inputs;

    char[] pversion;
    char[] name;
    char[] date;

    char[] filename;

    int opCmp(Object o) {
        ReplayEntry e = cast(ReplayEntry)o;
        return (allScore < e.allScore) - (allScore > e.allScore);
    }

    void addHeaderLine(char[] line, char[] errmsg) {
        char[][] kv = line.split(`: `);
        char[] k = kv[0];
        char[] v = join(kv[1..length], `: `);
        switch (k) {
        case `version`:
            pversion = v; break;
        case `score`:
            allScore = toInt(v); break;
        case `name`:
            name = v; break;
        case `date`:
            date = v; break;
        case `gameScore`:
            gameScore = toInt(v); break;
        case `turnScore`:
            turnScore = toInt(v); break;
        case `clearScore`:
            clearScore = toInt(v); break;
        case `leftScore`:
            leftScore = toInt(v); break;
        case `nomissScore`:
            nomissScore = toInt(v); break;
        case `turn`:
            endTurn = toInt(v); break;
        case `patterns`:
            patterns = toInt(v); break;
        case `seed`:
            seed = toLong(v); break;
        case `utcTime`:
            utcTime = toLong(v); break;
        case `mode`:
            mode = v; break;
        case `filename`:
            filename = v; break;
        case `inputEnc`:
            break;
        case `scoreStatEnc`:
            if (v == `divPrintable95`)
                scoreStatEnc = ScoreStatEnc.DIV_PRINTABLE95;
            break;
        case `scoreStat`:
            if (scoreStatEnc == ScoreStatEnc.DIV_PRINTABLE95) {
                int prev = 0;
                for (int i = 0; i < v.length; i += 4) {
                    int s = prev +
                        (v[i+0]-32) +
                        (v[i+1]-32)*95 +
                        (v[i+2]-32)*95*95 +
                        (v[i+3]-32)*95*95*95;
                    scoreStat ~= s;
                    prev = s;
                }
            }
            else {
                foreach (char[] s; v.split(`,`)) {
                    if (s.length == 0) break;
                    scoreStat ~= toInt(s);
                }
            }
            break;
        default:
            Log.warn(k ~ `: unknown key in replay ` ~ errmsg);
        }
    }
}

abstract class ReplayDB {
    void setSort(char[] mode, SceneSelect scene) {
        sorted.length = 0;
        foreach (ReplayEntry e; entries) {
            if (e.mode == mode) sorted ~= e;
        }
        sorted.sort;

        scene.setResultDisplay();
    }

    void prepareReplayFile(ReplayEntry entry) {}

    bool isOk() { return ok_; }
    bool isError() { return error_; }

    ReplayEntry[] entries;
    ReplayEntry[] sorted;

    bool ok_;
    bool error_;

}

private import std.stream;
private import std.file;
private import std.regexp;

class LocalReplayDB : ReplayDB {
    int load() {
        RegExp rpyReg =
            new RegExp(`^\d\d\d\d\d\d\d\d\d\d\d\d_[a-zA-Z0-9%]+\.rpy$`, ``);
        foreach (char[] fname; listdir(`replay`)) {
            if (!rpyReg.test(fname)) continue;

            try {
                ReplayEntry entry = new ReplayEntry();

                File ifile = new File(`replay/` ~ fname);
                char[] line;
                while ((line = ifile.readLine).length != 0) {
                    entry.addHeaderLine(line, fname);
                }

                ifile.close();

                entries ~= entry;
            }
            catch (Object o) {
                Log.warn(fname~": broken replay file ("~o.toString()~").");
            }
        }

        ok_ = true;

        error_ = entries.length == 0;

        return 0;
    }

    this() {
        ok_ = false;
        thread_ = new Thread(&load);
        thread_.start();
    }

    Thread thread_;
}

private import http;
private import sino;
private import rc;

class NetReplayDB : ReplayDB {
    this() {
        ok_ = true;
    }

    override void setSort(char[] mode, SceneSelect scene) {
        mode_ = mode;
        ok_ = false;
        scene_ = scene;
        thread_ = new Thread(&load);
        thread_.start();
    }

    int load() {
        try {
            char[] host = Sino.rc.get(`scoreServer`);
            int port = Sino.rc.getInt(`scorePort`);
            char[] dir = Sino.rc.get(`scorePath`);

            Http h = new Http10(host, port);
            char[] path = dir ~ `/scores_` ~ mode_ ~ `.txt`;
            HttpResponse res = h.get(path);

            Log.info(`score server status: ` ~ res.statusLine);

            sorted.length = 0;
            ReplayEntry e = new ReplayEntry();
            foreach (char[] line; res.messageBody.split("\n")) {
                if (line.length > 0) {
                    e.addHeaderLine(line, path);
                }
                else {
                    if (e.name) {
                        sorted ~= e;
                        e = new ReplayEntry();
                    }
                }
            }

            sorted.sort;

            error_ = false;

            scene_.setResultDisplay();
        }
        catch (Object o) {
            Log.info(`cannot connect score server: ` ~ o.toString());
            error_ = true;
        }

        ok_ = true;

        return 0;
    }

    override void prepareReplayFile(ReplayEntry entry) {
        ok_ = false;
        entry_ = entry;
        thread_ = new Thread(&prepare);
        thread_.start();
    }

    private int prepare() {
        try {
            char[] rpyfile = `replay/` ~ entry_.filename;
            if (std.file.exists(rpyfile)) {
                ok_ = true;
                return 0;
            }

            Log.info(rpyfile ~ `: get from score server`);

            char[] host = Sino.rc.get(`scoreServer`);
            int port = Sino.rc.getInt(`scorePort`);
            char[] dir = Sino.rc.get(`scorePath`);

            Http h = new Http10(host, port);
            char[] path = dir ~ `/` ~ rpyfile;
            path = path.replace(`%`, `%25`);
            HttpResponse res = h.get(path);

            char[] rpy = res.messageBody;
            std.file.write(rpyfile, rpy);
        }
        catch (Object o) {
            Log.info(`cannot get replay file: ` ~ o.toString());
            error_ = true;
        }

        ok_ = true;

        return 0;
    }

private:
    Thread thread_;
    char[] mode_;
    SceneSelect scene_;
    ReplayEntry entry_;
}
